package main

import (
	"fmt"
	"os"

	"git.sr.ht/~rjarry/go-opt/v2"
	"github.com/emersion/go-message/mail"
)

var (
	Version string = "v1.0.1"
	Date    string = "1970-01-01"
)

type args struct {
	Help    bool `opt:"-h,--help" action:"ShowHelp"`
	Version bool `opt:"--version" action:"PrintVersion"`
	// mjmap specific flags
	Mailboxes []string `opt:"--mailbox" action:"ParseMailbox" metavar:"<name>"`
	// msmtp compatibility flags
	ServerInfo bool `opt:"-S,--serverinfo"`
	Debug      bool `opt:"-d,--debug" action:"SetDebug"`
	// sendmail compatibility flags
	ReadRecipients bool     `opt:"-t,--read-recipients"`
	From           string   `opt:"-f,--from"`
	IgnoreDots     bool     `opt:"-i,--ignore-dots"`
	LogFile        string   `opt:"-X,--logfile" metavar:"<file>"`
	Recipients     []string `opt:"..." action:"ParseRecipient" required:"false" metavar:"<recipient...>"`
}

func (a *args) ShowHelp(string) error {
	fmt.Printf("Usage: %s\n", opt.NewCmdSpec(os.Args[0], a).Usage())
	fmt.Print(`
mjmap is a sendmail-compatible JMAP client.

Positional arguments:

 <recipient...>          Email address of a recipient to send the message to.
                         Can be specified multiple times.

Options:

  -h, --help             Show this help message and exit.
  --version              Print version information.
  --mailbox=<name>       Additional mailbox to add the message to, in addition
                         to Sent. Can be specified multiple times.
  -S, --server-info      Print information about the JMAP server and exit.
  -d, --debug            Print lots of debugging information.
  -t, --read-recipients  Read recipient addresses from the To, Cc, and Bcc
                         headers of the mail in addition of the ones specified
                         on the command line.
  -f, --from=<from>      Set the envelope-from address if it is missing from
                         the email headers.
  -i, --ignore-dots      Compatibility with sendmail, ignored option.
  -X, --logfile=<file>   Compatibility with sendmail, ignored option.
`)
	os.Exit(1)
	return nil
}

func (a *args) ParseMailbox(arg string) error {
	a.Mailboxes = append(a.Mailboxes, arg)
	return nil
}

func (a *args) PrintVersion(string) error {
	fmt.Printf("mjmap %s (%s)\n", Version, Date)
	os.Exit(1)
	return nil
}

func (a *args) SetDebug(string) error {
	DebugInit()
	return nil
}

func (a *args) ParseRecipient(arg string) error {
	if addr, err := mail.ParseAddress(arg); err != nil {
		return fmt.Errorf("%q: %w", arg, err)
	} else {
		a.Recipients = append(a.Recipients, addr.Address)
	}
	return nil
}

func main() {
	var args args

	spec := opt.NewCmdSpec(os.Args[0], &args)

	if err := spec.ParseArgs(opt.QuoteArgs(os.Args...)); err != nil {
		fmt.Fprintf(os.Stderr, "error: %s\n", err)
		fmt.Fprintf(os.Stderr, "Usage: %s\n", spec.Usage())
		os.Exit(1)
	}

	if err := mjmap(&args); err != nil {
		fmt.Fprintf(os.Stderr, "error: %s\n", err)
		os.Exit(1)
	}
}
