package main

import (
	"bufio"
	"bytes"
	"fmt"
	"io"
	"os"
	"strings"
	"sync"

	"git.sr.ht/~rockorager/go-jmap"
	"git.sr.ht/~rockorager/go-jmap/core"
	"git.sr.ht/~rockorager/go-jmap/mail"
	"git.sr.ht/~rockorager/go-jmap/mail/email"
	"git.sr.ht/~rockorager/go-jmap/mail/emailsubmission"
	"git.sr.ht/~rockorager/go-jmap/mail/identity"
	"git.sr.ht/~rockorager/go-jmap/mail/mailbox"
	"github.com/emersion/go-message"
	msgmail "github.com/emersion/go-message/mail"
)

func mjmap(args *args) error {
	acct, err := loadConfig()
	if err != nil {
		return err
	}

	client := &jmap.Client{
		SessionEndpoint: acct.endpoint,
	}

	if acct.token != "" {
		client.WithAccessToken(acct.token)
	} else if acct.password != "" {
		client.WithBasicAuth(acct.username, acct.token)
	}

	// TODO cache Session object
	if err := client.Authenticate(); err != nil {
		return err
	}

	if args.ServerInfo {
		return serverInfo(client.Session)
	}

	acctID, ok := client.Session.PrimaryAccounts[mail.URI]
	if !ok {
		return fmt.Errorf("no primary mail account")
	}

	msg, err := message.Read(crlf(os.Stdin))
	if err != nil {
		return fmt.Errorf("failed to parse email message: %w", err)
	}
	if args.From != "" && !msg.Header.Has("From") {
		msg.Header.Set("From", args.From)
	}
	from, err := msgmail.ParseAddress(msg.Header.Get("From"))
	if err != nil {
		return fmt.Errorf("failed to parse From address: %w", err)
	}

	// always use recipients from command line
	recipients := make(map[string]struct{})
	for _, r := range args.Recipients {
		recipients[r] = struct{}{}
	}
	if args.ReadRecipients {
		// add recipients from the email envelope
		for _, header := range []string{"To", "Cc", "Bcc"} {
			value := strings.TrimSpace(msg.Header.Get(header))
			if value == "" {
				continue
			}
			addrs, err := msgmail.ParseAddressList(value)
			if err != nil {
				return fmt.Errorf("failed to parse %s header: %w", header, err)
			}
			for _, a := range addrs {
				recipients[a.Address] = struct{}{}
			}
		}
	}
	if len(recipients) == 0 {
		return fmt.Errorf("no recipients for email")
	}
	// ensure not to leak the bcc header, if any
	msg.Header.Del("Bcc")

	// dump the (possibly modified) message to a byte buffer
	var buf bytes.Buffer
	if err := msg.WriteTo(&buf); err != nil {
		return fmt.Errorf("failed to reformat message: %w", err)
	}

	// Upload the data
	wg := sync.WaitGroup{}
	var (
		blobID    jmap.ID
		uploadErr error
	)
	wg.Add(1)
	go func() {
		defer wg.Done()
		uploadResp, err := client.Upload(acctID, &buf)
		if err != nil {
			uploadErr = err
			return
		}
		Debugf("Upload Success blobID=%s size=%d", uploadResp.ID, uploadResp.Size)
		blobID = uploadResp.ID
	}()

	req := &jmap.Request{}

	// Get the name, role, and ID (default) of all mailboxes. We'll maybe
	// need mailbox names if the user added an extra mailbox flag
	req.Invoke(&mailbox.Get{
		Account: acctID,
		Properties: []string{
			"name",
			"role",
		},
	})

	req.Invoke(&identity.Get{
		Account: acctID,
	})

	mboxResp, err := client.Do(req)
	if err != nil {
		return err
	}
	var (
		drafts  jmap.ID
		sent    jmap.ID
		identID jmap.ID
		extras  []string
	)
	for _, r := range mboxResp.Responses {
		switch r := r.Args.(type) {
		case *identity.GetResponse:
			name, domain, _ := strings.Cut(from.Address, "@")
			for _, id := range r.List {
				n, d, _ := strings.Cut(id.Email, "@")
				switch {
				case n == name && d == domain:
					fallthrough
				case n == "*" && d == domain:
					identID = id.ID
				}
			}
		case *mailbox.GetResponse:
			for _, mbox := range r.List {
				if mbox.Role == mailbox.RoleDrafts {
					drafts = mbox.ID
					continue
				}
				if mbox.Role == mailbox.RoleSent {
					sent = mbox.ID
					continue
				}
				for _, extra := range args.Mailboxes {
					if mbox.Name == extra {
						Debugf("%s mailbox ID=%s", mbox.Name, mbox.ID)
						extras = append(extras, string(mbox.ID))
					}
				}
			}
		}
	}

	if identID == "" {
		return fmt.Errorf("not allowed to send email as From: %s", from.Address)
	}
	if drafts == "" {
		return fmt.Errorf("no drafts mailbox found")
	}
	if sent == "" {
		return fmt.Errorf("no sent mailbox found")
	}
	Debugf("Drafts mailbox ID=%s", drafts)
	Debugf("Sent mailbox ID=%s", sent)

	wg.Wait()
	if uploadErr != nil {
		return uploadErr
	}

	// Import the blob into drafts
	imprt := &email.EmailImport{
		BlobID:     blobID,
		MailboxIDs: make(map[jmap.ID]bool),
		Keywords: map[string]bool{
			"$draft": true,
			"$seen":  true,
		},
	}
	imprt.MailboxIDs[drafts] = true
	req = &jmap.Request{}
	req.Invoke(&email.Import{
		Account: acctID,
		Emails: map[string]*email.EmailImport{
			"mjmap": imprt,
		},
	})

	Debugf("From: %s", from)
	var rcpTo []*emailsubmission.Address
	for r := range recipients {
		Debugf("Recipient: %s", r)
		rcpTo = append(rcpTo, &emailsubmission.Address{Email: r})
	}

	// Create the submission
	submission := &emailsubmission.EmailSubmission{
		IdentityID: identID,
		Envelope: &emailsubmission.Envelope{
			MailFrom: &emailsubmission.Address{Email: from.Address},
			RcptTo:   rcpTo,
		},
		EmailID: "#mjmap",
	}
	sentPath := fmt.Sprintf("mailboxIds/%s", sent)
	draftPath := fmt.Sprintf("mailboxIds/%s", drafts)
	patch := jmap.Patch{
		"keywords/$draft": nil,
		sentPath:          true,
		draftPath:         nil,
	}
	for _, id := range extras {
		path := fmt.Sprintf("mailboxIds/%s", id)
		patch[path] = true
	}
	req.Invoke(&emailsubmission.Set{
		Account: acctID,
		Create: map[jmap.ID]*emailsubmission.EmailSubmission{
			"id1": submission,
		},
		OnSuccessUpdateEmail: map[jmap.ID]jmap.Patch{
			"#id1": patch,
		},
	})

	importResp, err := client.Do(req)
	if err != nil {
		return err
	}
	for _, r := range importResp.Responses {
		switch r := r.Args.(type) {
		case *jmap.MethodError:
			return r
		case *email.ImportResponse:
			for _, v := range r.Created {
				Debugf("Created Import ID=%s", v.ID)
			}
			for _, v := range r.NotCreated {
				return fmt.Errorf("email not created: %v", v.Type)
			}
		case *emailsubmission.SetResponse:
			for _, v := range r.Created {
				Debugf("Created submission ID=%s", v.ID)
			}
			for _, v := range r.NotCreated {
				return fmt.Errorf("Submission not created: %v", v.Type)
			}
		}
	}

	return nil
}

// Prints the server info
func serverInfo(s *jmap.Session) error {
	fmt.Println("Capabilities")
	for _, capability := range s.Capabilities {
		switch cap := capability.(type) {
		case *core.Core:
			fmt.Printf("    Max Upload Size=%dMB\n", cap.MaxSizeUpload/1e6)
		}
	}

	fmt.Println("\nAccounts")
	for id, acct := range s.Accounts {
		cap, ok := acct.Capabilities[mail.URI]
		if !ok {
			continue
		}
		mailCap := cap.(*mail.Mail)
		cap, ok = acct.Capabilities[emailsubmission.URI]
		if !ok {
			continue
		}
		submissionCap := cap.(*emailsubmission.Capability)
		fmt.Printf("    %s", acct.Name)
		if id == s.PrimaryAccounts[mail.URI] {
			fmt.Printf(" (primary)")
		}
		fmt.Println()
		fmt.Printf("        Max Size Attachment per email=%dMB\n", mailCap.MaxSizeAttachmentsPerEmail/1e6)
		fmt.Printf("        Max Delayed Send=%d seconds\n", submissionCap.MaxDelayedSend)
		fmt.Println()
	}

	return nil
}

// crlf wraps an io.Reader with CRLF terminated lines
func crlf(r io.Reader) io.Reader {
	var buf bytes.Buffer
	scanner := bufio.NewScanner(r)
	for scanner.Scan() {
		buf.WriteString(scanner.Text() + "\r\n")
	}
	return &buf
}
